
/*
 * Types/enumerations for [Atom] DBO (Dynamic binary optimization)
 * investigation programs.
 *
 * Copyright (C) 1999,2000
 *	Rick Gorton	gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <machine/inst.h>

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

typedef unsigned char	U8;
typedef unsigned char	u8;
typedef u8	BOOLEAN;


#define	SLL_LIT_OPCODE		0x48001720
#define	SLL_REG_OPCODE		0x48000720

#define	SRL_LIT_OPCODE		0x48001680
#define	SRL_REG_OPCODE		0x48000680

#define	SRA_LIT_OPCODE		0x48001780
#define	SRA_REG_OPCODE		0x48000780

#define	FBEQ_OPCODE		0xc4000000
#define	FBGE_OPCODE		0xd8000000
#define	FBGT_OPCODE		0xdc000000
#define	FBLE_OPCODE		0xcc000000
#define	FBLT_OPCODE		0xc8000000
#define	FBNE_OPCODE		0xd4000000

#define	FTOIS_OPCODE		0x701f0f00
#define	FTOIT_OPCODE		0x701f0e00

#define	ADDS_OPCODE		0x58001000
#define	ADDT_OPCODE		0x58001400

#define	SUBS_OPCODE		0x58001020
#define	SUBT_OPCODE		0x58001420

#define	SQRTS_OPCODE		0x53e01160
#define	SQRTT_OPCODE		0x53e01560

#define	MULS_OPCODE		0x58001040
#define	MULT_OPCODE		0x58001440

#define	CVTLQ_OPCODE		0x5fe00200
#define	CVTTQ_OPCODE		0x5be015e0
#define	CVTTS_OPCODE		0x5be01580
#define	CVTQL_OPCODE		0x5fe00600
#define	CVTST_OPCODE		0x5be05580

#define	DIVS_OPCODE		0x58001060
#define	DIVT_OPCODE		0x58001460

#define	OPCODE_END		0xffff
#define	FUNCTION_CODE_END	0xffff

enum Opcodes {
	SLL_LIT_OP,
	SLL_REG_OP,
	SRL_LIT_OP,
	SRL_REG_OP,
	SRA_LIT_OP,
	SRA_REG_OP,
	FBEQ_OP,
	FBGE_OP,
	FBGT_OP,
	FBLE_OP,
	FBLT_OP,
	FBNE_OP,
	FTOIS_OP,
	FTOIT_OP,
	ADDS_OP,
	ADDT_OP,
	SUBS_OP,
	SUBT_OP,
	SQRTS_OP,
	SQRTT_OP,
	MULS_OP,
	MULT_OP,
	CVTLQ_OP,
	CVTTQ_OP,
	CVTTS_OP,
	CVTQL_OP,
	CVTST_OP,
	DIVS_OP,
	DIVT_OP,
	LAST_OP
	};

static const int Opcode_Value[(int) OPCODE_END] = {
	SLL_LIT_OPCODE,
	SLL_REG_OPCODE,
	SRL_LIT_OPCODE,
	SRL_REG_OPCODE,
	SRA_LIT_OPCODE,
	SRA_REG_OPCODE,
	FBEQ_OPCODE,
	FBGE_OPCODE,
	FBGT_OPCODE,
	FBLE_OPCODE,
	FBLT_OPCODE,
	FBNE_OPCODE,
	FTOIS_OPCODE,
	FTOIT_OPCODE,
	ADDS_OPCODE,
	ADDT_OPCODE,
	SUBS_OPCODE,
	SUBT_OPCODE,
	SQRTS_OPCODE,
	SQRTT_OPCODE,
	MULS_OPCODE,
	MULT_OPCODE,
	CVTLQ_OPCODE,
	CVTTQ_OPCODE,
	CVTTS_OPCODE,
	CVTQL_OPCODE,
	CVTST_OPCODE,
	DIVS_OPCODE,
	DIVT_OPCODE,
	OPCODE_END
	};

static const char Opcode_String[(int) OPCODE_END][32] = {
	"SLL_L",
	"SLL_R",
	"SRL_L",
	"SRL_R",
	"SRA_L",
	"SRA_R",
	"FBEQ",
	"FBGE",
	"FBGT",
	"FBLE",
	"FBLT",
	"FBNE",
	"FTOIS",
	"FTOIT",
	"ADDS",
	"ADDT",
	"SUBS",
	"SUBT",
	"SQRTS",
	"SQRTT",
	"MULS",
	"MULT",
	"CVTLQ",
	"CVTTQ",
	"CVTTS",
	"CVTQL",
	"CVTST",
	"DIVS",
	"DIVT",
	"CONSISTENCY ERROR"
	};

typedef union {
		long	longstuff;
		double	doublestuff;
		struct {
			long	mantissa:52;
			long	exponent:11;
			long	sign:1;
			} rawbits;
		} IEEE_UNION_;

typedef struct {
	long	counter;
	long	location;
	long	this_opnd[3];	/* 3 possible operands */
	} INSN_TYPE;
/*
 * Structures for Atom-izing time data collection
 */
static long static_count[(int)LAST_OP];
INSN_TYPE static_weird_FNOP[(int)LAST_OP];
INSN_TYPE static_zero_opnd[(int)LAST_OP];
INSN_TYPE static_float_copy[(int)LAST_OP];
INSN_TYPE static_float_clear[(int)LAST_OP];
INSN_TYPE static_invariant_float_branch[(int)LAST_OP];

/*
 * Structures for execution time data collection
 */
long dyn_count[(int)LAST_OP];
INSN_TYPE dyn_weird_FNOP[(int)LAST_OP];
INSN_TYPE dyn_zero_opnd[(int)LAST_OP];
INSN_TYPE dyn_dual_zero[(int)LAST_OP];
INSN_TYPE dyn_float_copy[(int)LAST_OP];
INSN_TYPE dyn_invariant_float_branch[(int)LAST_OP];
INSN_TYPE dyn_float_clear[(int)LAST_OP];

FILE	*Inst_Logfile;

/*
 * Macros (YUK) and constants for extracting the LIT field from an
 * operate instruction, since the ATOM people didn't deem it important
 * enough to provide this feature.  (Grumble)
 */

#define	OPERATE_MASK		0xfc001fe0
#define LITERAL_REGISTER_MASK	0x1000
#define LITERAL_FIELD_MASK	0x1fe000
#define LITERAL_SHIFT_COUNT	13
#define LITERAL_BIT_MASK	0xff

#define IS_LITERAL_OPERATE_(OPCODE32)	\
	(OPCODE32 & LITERAL_REGISTER_MASK)
#define LITERAL_VALUE_(OPCODE32)	\
	((OPCODE32 >> LITERAL_SHIFT_COUNT) & LITERAL_BIT_MASK)

