
/*
 * routines to collect dynamic data about shift instructions
 */

/*
 * Copyright (C) 2000   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>
#include "datatypes.h"

/*
 * Data declarations local to this file
 */
#define	MINUS_PI	-3.14150
#define	CACHE_TABLE_SIZE_	1

typedef struct {
	double	opnd1_val;
	double	opnd2_val;	/* Not used by SQRTx */
	long	hits;
	long	prior_address;
	long	this_address;
	} FP_VALUE_CACHE_;

typedef struct {
	long	mantissa_zero_count;
	long	neg_count;	/* -1.0 count for MULx, DIVx */
				/* Used as count of < 0 inputs for SQRTx */
	long	zero_count;
	long	one_count;
	} FP_EVAL_;

FP_EVAL_	MULx_count;
FP_VALUE_CACHE_	MULx_cache[CACHE_TABLE_SIZE_];

FP_EVAL_	DIVx_count;
FP_VALUE_CACHE_	DIVx_cache[CACHE_TABLE_SIZE_];
long		DIVx_opnds_equal;
long		DIVx_one_div_one;
long		DIVx_inverse_count;

FP_EVAL_	SQRTx_count;
FP_VALUE_CACHE_	SQRTx_cache[CACHE_TABLE_SIZE_];

/*
 * Initialization
 */
void DynamicApplicationInit()
{
  int	i;

  for (i=0; i<= (int)LAST_OP; i++)
    {
    dyn_count[i] = 0x0;

    dyn_weird_FNOP[i].counter = 0x0;
    dyn_weird_FNOP[i].location = 0x0;
    dyn_weird_FNOP[i].this_opnd[0] = 0x0;
    dyn_weird_FNOP[i].this_opnd[1] = 0x0;
    dyn_weird_FNOP[i].this_opnd[2] = 0x0;

    dyn_zero_opnd[i].counter = 0x0;
    dyn_zero_opnd[i].location = 0x0;
    dyn_zero_opnd[i].this_opnd[0] = 0x0;
    dyn_zero_opnd[i].this_opnd[1] = 0x0;
    dyn_zero_opnd[i].this_opnd[2] = 0x0;

    dyn_dual_zero[i].counter = 0x0;
    dyn_dual_zero[i].location = 0x0;
    dyn_dual_zero[i].this_opnd[0] = 0x0;
    dyn_dual_zero[i].this_opnd[1] = 0x0;
    dyn_dual_zero[i].this_opnd[2] = 0x0;

    dyn_float_copy[i].counter = 0x0;
    dyn_float_copy[i].location = 0x0;
    dyn_float_copy[i].this_opnd[0] = 0x0;
    dyn_float_copy[i].this_opnd[1] = 0x0;
    dyn_float_copy[i].this_opnd[2] = 0x0;

    dyn_invariant_float_branch[i].counter = 0x0;
    dyn_invariant_float_branch[i].location = 0x0;
    dyn_invariant_float_branch[i].this_opnd[0] = 0x0;
    dyn_invariant_float_branch[i].this_opnd[1] = 0x0;
    dyn_invariant_float_branch[i].this_opnd[2] = 0x0;

    dyn_float_clear[i].counter = 0x0;
    dyn_float_clear[i].location = 0x0;
    dyn_float_clear[i].this_opnd[0] = 0x0;
    dyn_float_clear[i].this_opnd[1] = 0x0;
    dyn_float_clear[i].this_opnd[2] = 0x0;
    }

  MULx_count.mantissa_zero_count = 0x0;
  MULx_count.neg_count = 0x0;
  MULx_count.zero_count = 0x0;
  MULx_count.one_count = 0x0;

  DIVx_count.mantissa_zero_count = 0x0;
  DIVx_count.neg_count = 0x0;
  DIVx_count.zero_count = 0x0;
  DIVx_count.one_count = 0x0;
  DIVx_opnds_equal = 0x0;
  DIVx_one_div_one = 0x0;
  DIVx_inverse_count = 0x0;

  SQRTx_count.mantissa_zero_count = 0x0;
  SQRTx_count.neg_count = 0x0;
  SQRTx_count.zero_count = 0x0;
  SQRTx_count.one_count = 0x0;

  for (i = 0; i < CACHE_TABLE_SIZE_; i ++)
    {
    MULx_cache[i].opnd1_val = MINUS_PI;
    MULx_cache[i].opnd2_val = MINUS_PI;
    MULx_cache[i].hits = 0x0;
    MULx_cache[i].prior_address = 0x0;
    MULx_cache[i].this_address = 0x0;

    DIVx_cache[i].opnd1_val = MINUS_PI;
    DIVx_cache[i].opnd2_val = MINUS_PI;
    DIVx_cache[i].hits = 0x0;
    DIVx_cache[i].prior_address = 0x0;
    DIVx_cache[i].this_address = 0x0;

    SQRTx_cache[i].opnd1_val = MINUS_PI;
    SQRTx_cache[i].opnd2_val = MINUS_PI;	/* Not relevant */
    SQRTx_cache[i].hits = 0x0;
    SQRTx_cache[i].prior_address = 0x0;
    SQRTx_cache[i].this_address = 0x0;
    }
}

/*
 * Print out results
 */
void DumpDynamicData()
{
  int   i;
  long	netval;
  double	tmpval;

  fprintf(Inst_Logfile, "*** dyn data begin ***\n");
  fprintf(Inst_Logfile, "    Weird FNOPs (using f31 as RegC)\n");
  for (i = 0; i < (int) LAST_OP; i++)
    {
    if (dyn_weird_FNOP[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%ld\t(one such at 0x%lx)\n",
			Opcode_String[i],
			dyn_weird_FNOP[i].counter,
			dyn_weird_FNOP[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Dual Zero Operands (RegA and RegB == 0x0)\n");
  for (i = 0; i < (int) LAST_OP; i++)
    {
    if (dyn_dual_zero[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%ld\t(one such at 0x%lx)\n",
			Opcode_String[i],
			dyn_dual_zero[i].counter,
			dyn_dual_zero[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Zero Operands (RegA or RegB == 0x0)\n");
  for (i = 0; i < (int) LAST_OP; i++)
    {
    if (dyn_zero_opnd[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%ld\t(one such at 0x%lx)\n",
			Opcode_String[i],
			dyn_zero_opnd[i].counter,
			dyn_zero_opnd[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Float copies (other than cpys) using f31\n");
  for (i = 0; i < (int) LAST_OP; i++)
    {
    if (dyn_float_copy[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%ld\t(one such at 0x%lx)\n",
			Opcode_String[i],
			dyn_float_copy[i].counter,
			dyn_float_copy[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Float clears (other than cpys)\n");
  for (i = 0; i < (int) LAST_OP; i++)
    {
    if (dyn_float_clear[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%ld\t(one such at 0x%lx)\n",
			Opcode_String[i],
			dyn_float_clear[i].counter,
			dyn_float_clear[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Invariant float branches (f31)\n");
  for (i = 0; i < (int) LAST_OP; i++)
    {
    if (dyn_invariant_float_branch[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%ld\t(one such at 0x%lx)\n",
			Opcode_String[i],
			dyn_invariant_float_branch[i].counter,
			dyn_invariant_float_branch[i].location);
	}
    }

	/*
	 * The important data (the rest is basically fluff)
	 * NOTE: 50.0 is used to represent 100.0% / 2 operands.
	 */

	/* ADDx */
  tmpval = dyn_count[ADDS_OP] + dyn_count[ADDT_OP];
  if (tmpval == 0.0e0)
    tmpval = 1.0e0;
  fprintf(Inst_Logfile, "\n\tADDx instruction data\n");
  fprintf(Inst_Logfile, "    Number ADDS uses:\t\t\t\t%ld\n", dyn_count[ADDS_OP]);
  fprintf(Inst_Logfile, "    Number ADDT uses:\t\t\t\t%ld\n", dyn_count[ADDT_OP]);
  fprintf(Inst_Logfile, "    Number ADDx operands:\t\t\t%ld\n",
			2 * (dyn_count[ADDT_OP] + dyn_count[ADDS_OP]));
  netval = dyn_float_clear[ADDS_OP].counter + dyn_float_clear[ADDT_OP].counter;
  fprintf(Inst_Logfile, "    Number ADDx X + -X:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) (netval) / tmpval);
  netval = dyn_dual_zero[ADDS_OP].counter + dyn_dual_zero[ADDT_OP].counter;
  fprintf(Inst_Logfile, "    Number ADDx 0.0 + 0.0:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) (netval) / tmpval);
  netval = dyn_zero_opnd[ADDS_OP].counter + dyn_zero_opnd[ADDT_OP].counter;
  fprintf(Inst_Logfile, "    Number ADDx X + Y where X or Y == 0.0:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) (netval) / tmpval);

	/* SUBx */
  tmpval = dyn_count[SUBS_OP] + dyn_count[SUBT_OP];
  if (tmpval == 0.0e0)
    tmpval = 1.0e0;
  fprintf(Inst_Logfile, "\n\tSUBx instruction data\n");
  fprintf(Inst_Logfile, "    Number SUBS uses:\t\t\t\t%ld\n", dyn_count[SUBS_OP]);
  fprintf(Inst_Logfile, "    Number SUBT uses:\t\t\t\t%ld\n", dyn_count[SUBT_OP]);
  netval = dyn_dual_zero[SUBS_OP].counter + dyn_dual_zero[SUBT_OP].counter;
  fprintf(Inst_Logfile, "    Number SUBx 0.0 - 0.0:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) (netval) / tmpval);
  netval = dyn_float_clear[SUBS_OP].counter + dyn_float_clear[SUBT_OP].counter;
  fprintf(Inst_Logfile, "    Number SUBx X - X:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) (netval) / tmpval);
  netval = dyn_zero_opnd[SUBS_OP].counter + dyn_zero_opnd[SUBT_OP].counter;
  fprintf(Inst_Logfile, "    Number SUBx X - 0:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) (netval) / tmpval);

	/* MULx */
  tmpval = dyn_count[MULS_OP] + dyn_count[MULT_OP];
  if (tmpval == 0.0e0)
    tmpval = 1.0e0;
  fprintf(Inst_Logfile, "\n\tMULx instruction data\n");
  fprintf(Inst_Logfile, "    Number MULS uses:\t\t\t\t%ld\n", dyn_count[MULS_OP]);
  fprintf(Inst_Logfile, "    Number MULT uses:\t\t\t\t%ld\n", dyn_count[MULT_OP]);
  fprintf(Inst_Logfile, "    Number MULx operands:\t\t\t\t%ld\n",
			2 * (dyn_count[MULT_OP] + dyn_count[MULS_OP]));
  fprintf(Inst_Logfile, "    Number MULx operands with mantissa of 0x0:\t%ld\t: %3.2f%%\n",
			MULx_count.mantissa_zero_count,
			(double) 50.0 * (double) MULx_count.mantissa_zero_count / tmpval);
  fprintf(Inst_Logfile, "    Number MULx operands with value -1.0:\t%ld\t: %3.2f%%\n",
			MULx_count.neg_count,
			(double) 50.0 * (double) MULx_count.neg_count / tmpval);
  netval = dyn_dual_zero[MULS_OP].counter + dyn_dual_zero[MULT_OP].counter;
  fprintf(Inst_Logfile, "    Number MULx 0.0 * 0.0:\t%ld\t: %3.2f%%\n",
			netval,
			(double) 100.0 * (double) netval / tmpval);
  fprintf(Inst_Logfile, "    Number MULx X * 0.0:\t%ld\t: %3.2f%%\n",
			MULx_count.zero_count,
			(double) 100.0 * (double) MULx_count.zero_count / tmpval);
  fprintf(Inst_Logfile, "    Number MULx operands with value 1.0:\t%ld\t: %3.2f%%\n",
			MULx_count.one_count,
			(double) 50.0 * (double) MULx_count.one_count / tmpval);
  netval = 0x0;
  for (i = 0; i < CACHE_TABLE_SIZE_; i++)
    netval = netval + MULx_cache[i].hits;
  fprintf(Inst_Logfile, "    MULx cache hits:\t\t%ld\n", netval);
  if (netval > 0)
    fprintf(Inst_Logfile, "    MULx cache hit rate:\t%2.2f%%\n",
		(double) (100 * netval)/(double) (dyn_count[MULS_OP] + dyn_count[MULT_OP]));
  else
    fprintf(Inst_Logfile, "    MULx cache hit rate:\t%2.2f%%\n", (double) netval);
  fprintf(Inst_Logfile, "    MULx operand cache breakout\n");
  fprintf(Inst_Logfile, "\tSlot\tHits\t\tRate (%%)\n");
  for (i = 0; i < CACHE_TABLE_SIZE_; i++) {
    if (netval > 0)
	fprintf(Inst_Logfile, "\t%ld\t%ld\t\t%2.2f\t@ 0x%lx :: 0x%lx\n", i, MULx_cache[i].hits,
		(double) (100 * MULx_cache[i].hits)/(double) (dyn_count[MULS_OP] + dyn_count[MULT_OP]),
		MULx_cache[i].prior_address, MULx_cache[i].this_address);
    else
	/* MULx_cache[i].hits is zero here */
	fprintf(Inst_Logfile, "\t%ld\t%ld\t\t%2.2f\n", i, 0, 0.0);
    }

	/* DIVx */
  tmpval = dyn_count[DIVS_OP] + dyn_count[DIVT_OP];
  if (tmpval == 0.0e0)
    tmpval = 1.0e0;
  fprintf(Inst_Logfile, "\n\tDIVx instruction data\n");
  fprintf(Inst_Logfile, "    Number DIVS uses:\t\t\t%ld\n", dyn_count[DIVS_OP]);
  fprintf(Inst_Logfile, "    Number DIVT uses:\t\t\t%ld\n", dyn_count[DIVT_OP]);
  fprintf(Inst_Logfile, "    Number of X / mantissa of 0x0:\t%ld\t: %3.2f%%\n",
			DIVx_count.mantissa_zero_count,
			(double) 100.0 * (double) DIVx_count.mantissa_zero_count / tmpval);
  fprintf(Inst_Logfile, "    Number of X / -1.0:\t\t\t%ld\t: %3.2f%%\n",
			DIVx_count.neg_count,
			(double) 100.0 * (double) DIVx_count.neg_count / tmpval);
  fprintf(Inst_Logfile, "    Number of 0.0 / Y:\t\t\t%ld\t: %3.2f%%\n",
			DIVx_count.zero_count,
			(double) 100.0 * (double) DIVx_count.zero_count / tmpval);
  fprintf(Inst_Logfile, "    Number of X / 1.0:\t\t\t%ld\t: %3.2f%%\n",
			DIVx_count.one_count,
			(double) 100.0 * (double) DIVx_count.one_count / tmpval);
  fprintf(Inst_Logfile, "    Number of X / X:\t\t\t%ld\t: %3.2f%%\n",
			DIVx_opnds_equal,
			(double) 100.0 * (double) DIVx_opnds_equal / tmpval);
  fprintf(Inst_Logfile, "    Number of 1.0 / 1.0:\t\t\t%ld\t: %3.2f%%\n",
			DIVx_one_div_one,
			(double) 100.0 * (double) DIVx_one_div_one / tmpval);
  fprintf(Inst_Logfile, "    Number of 1.0 / X:\t\t\t%ld\t: %3.2f%%\n",
			DIVx_inverse_count,
			(double) 100.0 * (double) DIVx_inverse_count / tmpval);
  netval = 0x0;
  for (i = 0; i < CACHE_TABLE_SIZE_; i++)
    netval = netval + DIVx_cache[i].hits;
  fprintf(Inst_Logfile, "    DIVx cache hits:\t\t%ld\n", netval);
  if (netval > 0)
    fprintf(Inst_Logfile, "    DIVx cache hit rate:\t%2.2f\n",
		(double) (100 * netval/(dyn_count[DIVS_OP] + dyn_count[DIVT_OP])));
  else
    fprintf(Inst_Logfile, "    DIVx cache hit rate:\t\t%2.2f\n", (double) netval);
  fprintf(Inst_Logfile, "    DIVx operand cache breakout\n");
  fprintf(Inst_Logfile, "\tSlot\tHits\t\tRate (%%)\n");
  for (i = 0; i < CACHE_TABLE_SIZE_; i++) {
    if (netval > 0)
	fprintf(Inst_Logfile, "\t%ld\t%ld\t\t%2.2f\t@ 0x%lx :: 0x%lx\n", i, DIVx_cache[i].hits,
		(double) (100 * DIVx_cache[i].hits/(dyn_count[DIVS_OP] + dyn_count[DIVT_OP])),
		DIVx_cache[i].prior_address, DIVx_cache[i].this_address);
    else
	/* DIVx_cache[i].hits is zero here */
	fprintf(Inst_Logfile, "\t%ld\t%ld\t\t%2.2f\n", i, 0, 0.0);
    }

	/* SQRTx */
  tmpval = dyn_count[DIVS_OP] + dyn_count[DIVT_OP];
  if (tmpval == 0.0e0)
    tmpval = 1.0e0;
  fprintf(Inst_Logfile, "\n\tSQRTx instruction data\n");
  fprintf(Inst_Logfile, "    Number SQRTS uses:\t\t\t%ld\n", dyn_count[SQRTS_OP]);
  fprintf(Inst_Logfile, "    Number SQRTT uses:\t\t\t%ld\n", dyn_count[SQRTT_OP]);
  fprintf(Inst_Logfile, "    Number SQRTx (mantissa of 0x0):\t%ld\t: %3.2f%%\n",
			SQRTx_count.mantissa_zero_count,
			(double) 50.0 * (double) SQRTx_count.mantissa_zero_count / tmpval);
  fprintf(Inst_Logfile, "    Number SQRTx (<0):\t\t\t%ld\t: %3.2f%%\n",
			SQRTx_count.neg_count,
			(double) 50.0 * (double) SQRTx_count.neg_count / tmpval);
  fprintf(Inst_Logfile, "    Number SQRTx (0.0):\t\t\t%ld\t: %3.2f%%\n",
			SQRTx_count.zero_count,
			(double) 50.0 * (double) SQRTx_count.zero_count / tmpval);
  fprintf(Inst_Logfile, "    Number SQRTx (1.0):\t\t\t%ld\t: %3.2f%%\n",
			SQRTx_count.one_count,
			(double) 50.0 * (double) SQRTx_count.one_count / tmpval);
  netval = 0x0;
  for (i = 0; i < CACHE_TABLE_SIZE_; i++)
    netval = netval + SQRTx_cache[i].hits;
  fprintf(Inst_Logfile, "    SQRTx cache hits:\t\t%ld\n", netval);
  if (netval > 0)
    fprintf(Inst_Logfile, "    SQRTx cache hit rate:\t%2.2f\n",
		(double) (100 * netval/(dyn_count[SQRTS_OP] + dyn_count[SQRTT_OP])));
  else
    fprintf(Inst_Logfile, "    SQRTx cache hit rate:\t%2.2f\n", (double) netval);
  fprintf(Inst_Logfile, "    SQRTx operand cache breakout\n");
  fprintf(Inst_Logfile, "\tSlot\tHits\t\tRate (%%)\n");
  for (i = 0; i < CACHE_TABLE_SIZE_; i++) {
    if (netval > 0)
	fprintf(Inst_Logfile, "\t%ld\t%ld\t\t%2.2f\t@ 0x%lx :: 0x%lx\n", i, SQRTx_cache[i].hits,
		(double) (100 * SQRTx_cache[i].hits/(dyn_count[SQRTS_OP] + dyn_count[SQRTT_OP])),
		SQRTx_cache[i].prior_address, SQRTx_cache[i].this_address);
    else
	/* SQRTx_cache[i].hits is zero here */
	fprintf(Inst_Logfile, "\t%ld\t%ld\t\t%2.2f\n", i, 0, 0.0);
    }

  fprintf(Inst_Logfile, "*** dyn data end ***\n");
}


/*
 * Instruction data collection
 */
void Weird_FNOP(int this_op, long this_PC)
{
  dyn_weird_FNOP[this_op].counter++;
  dyn_weird_FNOP[this_op].location = this_PC;
}

void Invariant_Float_Branch(int this_op, long this_PC)
{
  dyn_invariant_float_branch[this_op].counter++;
  dyn_invariant_float_branch[this_op].location = this_PC;
}

void Float_Copy(int this_op, long this_PC)
{
  dyn_float_copy[this_op].counter++;
  dyn_float_copy[this_op].location = this_PC;
}

void Check_Float_Zero_Operand(int this_op, long this_PC, int opnd_num, double this_opnd_value)
{
  IEEE_UNION_	op1;

  op1.doublestuff = this_opnd_value;
  if (op1.longstuff == 0x0)
    {
    dyn_zero_opnd[this_op].counter++;
    dyn_zero_opnd[this_op].location = this_PC;
    dyn_zero_opnd[this_op].this_opnd[opnd_num]++;
    }
}

static void Float_Dual_Zero(int this_op, long this_PC)
{
  dyn_dual_zero[this_op].counter++;
  dyn_dual_zero[this_op].location = this_PC;
}

void Float_Clear(int this_op, long this_PC)
{
  dyn_float_clear[this_op].counter++;
  dyn_float_clear[this_op].location = this_PC;
}

void Check_FADD(int this_op, long this_PC, double opnd_1, double opnd_2)
{
  IEEE_UNION_	op1, op2;

  dyn_count[this_op]++;
  if (opnd_1 == -opnd_2)
    Float_Clear(this_op, this_PC);
  else {
    op1.doublestuff = opnd_1;
    op2.doublestuff = opnd_2;

    if ((op1.longstuff == 0x0) && (op2.longstuff == 0x0))
	Float_Dual_Zero (this_op, this_PC);
    else if (op1.longstuff == 0x0)
	{
	dyn_zero_opnd[this_op].counter++;
	dyn_zero_opnd[this_op].location = this_PC;
	}
    else if (op2.longstuff == 0x0)
	{
	dyn_zero_opnd[this_op].counter++;
	dyn_zero_opnd[this_op].location = this_PC;
	}
    }
}

void Check_FSUB(int this_op, long this_PC, double opnd_1, double opnd_2)
{
  IEEE_UNION_	op1;

  dyn_count[this_op]++;
  if (opnd_1 == opnd_2)
    {
    op1.doublestuff = opnd_1;
    if (op1.longstuff != 0x0)
	Float_Clear(this_op, this_PC);
    else
	Float_Dual_Zero (this_op, this_PC);
    }
  else
    Check_Float_Zero_Operand(this_op, this_PC, 1, opnd_2);
}

void Check_FSQRT(int this_op, long this_PC, double opnd_1)
{
	/*
	 * Relevant unusual cases to look for:
	 *	operand < 0.0
	 *	operand of 0.0
	 *	operand of 1.0
	 *	mantissa of zero (including above)
	 * Also look up the operands in a value cache.
	 */
  IEEE_UNION_	op1;
  int		i;

  dyn_count[this_op]++;

  op1.doublestuff = opnd_1;
  if (op1.rawbits.mantissa == 0x0) {
    SQRTx_count.mantissa_zero_count++;
    if (opnd_1 < 0.0)
	SQRTx_count.neg_count++;
    else if (op1.longstuff == 0x0)
	SQRTx_count.zero_count++;
    else if (op1.rawbits.exponent == 0x3ff)
	SQRTx_count.one_count++;
    }
  for (i = 0; i < CACHE_TABLE_SIZE_; i++) {
    if (SQRTx_cache[i].opnd1_val == opnd_1)
	{
	/* Got a hit - record it and return */
	SQRTx_cache[i].hits++;
	SQRTx_cache[i].prior_address = SQRTx_cache[i].this_address;
	SQRTx_cache[i].this_address = this_PC;
	return;
	}
    }

	/*
	 * If we got here, we missed in the cache.  Age old values.
	 */
  for (i = CACHE_TABLE_SIZE_ - 1; i > 0; i--) {
    SQRTx_cache[i].opnd1_val = SQRTx_cache[i-1].opnd1_val;
    SQRTx_cache[i].prior_address = SQRTx_cache[i-1].prior_address;
    SQRTx_cache[i].this_address = SQRTx_cache[i-1].this_address;
    }

	/*
	 * New zero-eth cache value
	 */
  SQRTx_cache[0].opnd1_val = opnd_1;
  SQRTx_cache[0].prior_address = SQRTx_cache[0].this_address;
  SQRTx_cache[0].this_address = this_PC;
}

void Check_FMUL(int this_op, long this_PC, double opnd_1, double opnd_2)
{
	/*
	 * Relevant unusual cases to look for:
	 *	either operand -1.0
	 *	either operand 0.0
	 *	either operand 1.0
	 *	either operand has a mantissa of zero (including above)
	 * Also look up the operands in a value cache.
	 */
  IEEE_UNION_	op1, op2;
  int		i;

  dyn_count[this_op]++;

  op1.doublestuff = opnd_1;
  if (op1.rawbits.mantissa == 0x0) {
    MULx_count.mantissa_zero_count++;
    if (op1.rawbits.exponent == 0x3ff)
	{
	if (op1.rawbits.sign == 0x0)
	    MULx_count.one_count++;
	else
	    MULx_count.neg_count++;
	}
    }

  op2.doublestuff = opnd_2;
  if (op2.rawbits.mantissa == 0x0) {
    MULx_count.mantissa_zero_count++;
    if (op2.rawbits.exponent == 0x3ff)
	{
	if (op2.rawbits.sign == 0x0)
	    MULx_count.one_count++;
	else
	    MULx_count.neg_count++;
	}
    }

  if ((op1.longstuff == 0x0) && (op2.longstuff == 0x0))
    Float_Dual_Zero (this_op, this_PC);
  else if (op1.longstuff == 0x0)
    MULx_count.zero_count++;
  else if (op2.longstuff == 0x0)
    MULx_count.zero_count++;

  for (i = 0; i < CACHE_TABLE_SIZE_; i++) {
    if ((MULx_cache[i].opnd1_val == opnd_1)
	&& (MULx_cache[i].opnd2_val == opnd_2))
	{
	/* Got a hit - record it and return */
	MULx_cache[i].hits++;
	MULx_cache[i].prior_address = MULx_cache[i].this_address;
	MULx_cache[i].this_address = this_PC;
	return;
	}
    }

	/*
	 * If we got here, we missed in the cache.  Age old values.
	 */
  for (i = CACHE_TABLE_SIZE_ - 1; i > 0; i--) {
    MULx_cache[i].opnd1_val = MULx_cache[i-1].opnd1_val;
    MULx_cache[i].opnd2_val = MULx_cache[i-1].opnd2_val;
    MULx_cache[i].prior_address = MULx_cache[i-1].prior_address;
    MULx_cache[i].this_address = MULx_cache[i-1].this_address;
    }

	/*
	 * New zero-eth cache value
	 */
  MULx_cache[0].opnd1_val = opnd_1;
  MULx_cache[0].opnd2_val = opnd_2;
  MULx_cache[0].prior_address = MULx_cache[0].this_address;
  MULx_cache[0].this_address = this_PC;
  return;
}

void Check_FDIV(int this_op, long this_PC, double opnd_1, double opnd_2)
{
	/*
	 * Relevant unusual cases to look for:
	 *	X / -1.0
	 *	0.0 / Y
	 *	X / 1.0
	 *	X / X
	 *	1.0 / 1.0
	 *	X / mantissa of zero (including above)
	 * Also look up the operands in a value cache.
	 */
  IEEE_UNION_	op1, op2;
  int		i;

  dyn_count[this_op]++;
  op1.doublestuff = opnd_1;
  if (op1.longstuff == 0x0)
	DIVx_count.zero_count++;
  if ((op1.rawbits.mantissa == 0x0) && (op1.rawbits.exponent == 0x3ff))
	DIVx_inverse_count++;		/* Also gets -1.0/Y */

  op2.doublestuff = opnd_2;
  if (op2.rawbits.mantissa == 0x0) {
    DIVx_count.mantissa_zero_count++;
    if (op2.rawbits.exponent == 0x3ff)
	{
	if (op2.rawbits.sign == 0x0)
	    DIVx_count.one_count++;
	else
	    DIVx_count.neg_count++;
	}
    }

  if (op1.longstuff == op2.longstuff)
    {
    DIVx_opnds_equal++;

    if (op1.longstuff == 0x3ff0000000000000)
	DIVx_one_div_one++;
    }

  for (i = 0; i < CACHE_TABLE_SIZE_; i++) {
    if ((DIVx_cache[i].opnd1_val == opnd_1)
	&& (DIVx_cache[i].opnd2_val == opnd_2))
	{
	/* Got a hit - record it and return */
	DIVx_cache[i].hits++;
	DIVx_cache[i].prior_address = DIVx_cache[i].this_address;
	DIVx_cache[i].this_address = this_PC;
	return;
	}
    }

	/*
	 * If we got here, we missed in the cache.  Age old values.
	 */
  for (i = CACHE_TABLE_SIZE_ - 1; i > 0; i--) {
    DIVx_cache[i].opnd1_val = DIVx_cache[i-1].opnd1_val;
    DIVx_cache[i].opnd2_val = DIVx_cache[i-1].opnd2_val;
    DIVx_cache[i].prior_address = DIVx_cache[i-1].prior_address;
    DIVx_cache[i].this_address = DIVx_cache[i-1].this_address;
    }

	/*
	 * New zero-eth cache value
	 */
  DIVx_cache[0].opnd1_val = opnd_1;
  DIVx_cache[0].opnd2_val = opnd_2;
  DIVx_cache[0].prior_address = DIVx_cache[0].this_address;
  DIVx_cache[0].this_address = this_PC;
  return;
}
