
/*
 *	Atom program to count static and dynamic occurances of trivial
 *	floating point operations.  Dynamically check the operands for
 *	DIVx, SQRTx, and MULx to see if any are zero, one, or a power of two.
 *	The reason I'm making the tool look for power of two values is
 *	because of the frequency of powers of two that seem to be fed to
 *	sqrt().  This may also happen a lot in MULx and DIVx instructions.
 *
 *	Implementation HACK note: Check for REG_31 vs FREG_31, since we
 *	already know what instruction is being looked at, there isn't any
 *	real point in calling GetInstRegEnum()...
 */

/*
 * Copyright (C) 1999, 2000	Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"
#include "declare.h"

#define	INT_OPERATE_MASK	0x3f

void InstrumentInit(int iargc, char **iargv)
{
  int	i,j;
  char	namebuffer[256];

  if (iargc > 1)
    strcpy(namebuffer, iargv[1]);
  else
    strcpy(namebuffer, LOGFILE_NAME_);
  Inst_Logfile = fopen(namebuffer, "a+");
  printf("Logfile will be '%s'\n", namebuffer);
  fprintf(Inst_Logfile, "Logfile:\t%s\n", namebuffer);

  for (i=0; i< (int)LAST_OP; i++)
    {
    static_count[i] = 0;

    static_weird_FNOP[i].counter = 0x0;
    static_weird_FNOP[i].location = 0x0;
    static_weird_FNOP[i].this_opnd[0] = 0x0;
    static_weird_FNOP[i].this_opnd[1] = 0x0;
    static_weird_FNOP[i].this_opnd[2] = 0x0;

    static_float_copy[i].counter = 0x0;
    static_float_copy[i].location = 0x0;
    static_float_copy[i].this_opnd[0] = 0x0;
    static_float_copy[i].this_opnd[1] = 0x0;
    static_float_copy[i].this_opnd[2] = 0x0;

    static_zero_opnd[i].counter = 0x0;
    static_zero_opnd[i].location = 0x0;
    static_zero_opnd[i].this_opnd[0] = 0x0;
    static_zero_opnd[i].this_opnd[1] = 0x0;
    static_zero_opnd[i].this_opnd[2] = 0x0;

    static_invariant_float_branch[i].counter = 0x0;
    static_invariant_float_branch[i].location = 0x0;
    static_invariant_float_branch[i].this_opnd[0] = 0x0;
    static_invariant_float_branch[i].this_opnd[1] = 0x0;
    static_invariant_float_branch[i].this_opnd[2] = 0x0;

    static_float_clear[i].counter = 0x0;
    static_float_clear[i].location = 0x0;
    static_float_clear[i].this_opnd[0] = 0x0;
    static_float_clear[i].this_opnd[1] = 0x0;
    static_float_clear[i].this_opnd[2] = 0x0;
    }

  AddCallProto("Weird_FNOP(int, long)");
  AddCallProto("Invariant_Float_Branch(int, long)");
  AddCallProto("Float_Copy(int, long)");
  AddCallProto("Check_Float_Zero_Operand(int, long, int, FREGV)");
  AddCallProto("Float_Clear(int, long)");
  AddCallProto("Check_FSQRT(int, long, FREGV)");
  AddCallProto("Check_FMUL(int, long, FREGV, FREGV)");
  AddCallProto("Check_FDIV(int, long, FREGV, FREGV)");
  AddCallProto("Check_FSUB(int, long, FREGV, FREGV)");
  AddCallProto("Check_FADD(int, long, FREGV, FREGV)");

  AddCallProto("DynamicInit(char *)");
  AddCallProgram(ProgramBefore, "DynamicInit", namebuffer);
}

Instrument(int argc, char **argv, Obj *obj)
{
  Proc *p; Block *b;  Inst *this_inst;
  RegvType	RegA, RegB, RegC;
  enum Opcodes	curr_opcode;
  long		this_PC;

int idx;

  for (p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p)) {
    for (b = GetFirstBlock(p); b != NULL; b = GetNextBlock(b)) {
      for (this_inst = GetFirstInst(b); this_inst != NULL;
			this_inst = GetNextInst(this_inst)) {
	IClassType	this_class;

	this_class = GetInstClass(this_inst);
	RegA = GetInstRegEnum(this_inst, InstRA);
	RegB = GetInstRegEnum(this_inst, InstRB);
	RegC = GetInstRegEnum(this_inst, InstRC);
	curr_opcode = GetOpcode(this_inst);
	this_PC = InstPC(this_inst);

	if ((RegC == FREG_31) && (curr_opcode != LAST_OP))
		/* LAST_OP means that we didn't get a match */
	    {
	    static_weird_FNOP[curr_opcode].counter++;
	    static_weird_FNOP[curr_opcode].location = this_PC;
	    static_weird_FNOP[curr_opcode].this_opnd[2]++;
	    AddCallInst(this_inst, InstBefore, "Weird_FNOP", curr_opcode,
			this_PC);
	    }
	switch (curr_opcode)
	    {
	    case FBEQ_OP:
	    case FBGE_OP:
	    case FBGT_OP:
	    case FBLE_OP:
	    case FBLT_OP:
	    case FBNE_OP:
		{
		if (RegA == FREG_31)
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_invariant_float_branch[curr_opcode].counter++;
		    static_invariant_float_branch[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Invariant_Float_Branch",
				curr_opcode, this_PC);
		    }
		break;
		}

	    case FTOIS_OP:
	    case FTOIT_OP:
		{
		if (RegA == FREG_31)
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_copy[curr_opcode].counter++;
		    static_float_copy[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Copy",
				curr_opcode, this_PC);
		    break;
		    }
		AddCallInst(this_inst, InstBefore, "Check_Float_Zero_Operand",
				curr_opcode,
				this_PC,
				0,	/* Signifying InstRA */
				GetInstRegEnum(this_inst, InstRA));
		break;
		}

	    case ADDS_OP:
	    case ADDT_OP:
		{
		if ((RegA == FREG_31) || (RegB == FREG_31))
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_copy[curr_opcode].counter++;
		    static_float_copy[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Copy",
				curr_opcode, this_PC);
		    break;
		    }
		AddCallInst(this_inst, InstBefore, "Check_FADD",
				curr_opcode,
				this_PC,
				GetInstRegEnum(this_inst, InstRA),
				GetInstRegEnum(this_inst, InstRB));
		break;
		}

	    case SUBS_OP:
	    case SUBT_OP:
		{
		if (RegA == RegB)
		    {
		    static_float_clear[curr_opcode].counter++;
		    static_float_clear[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Clear",
				curr_opcode, this_PC);
		    break;
		    }
		if (RegB == REG_31)
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_copy[curr_opcode].counter++;
		    static_float_copy[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Copy",
				curr_opcode, this_PC);
		    break;
		    }
		AddCallInst(this_inst, InstBefore, "Check_FSUB",
				curr_opcode,
				this_PC,
				GetInstRegEnum(this_inst, InstRA),
				GetInstRegEnum(this_inst, InstRB));
		break;
		}

	    case SQRTS_OP:
	    case SQRTT_OP:
		{
		static_count[curr_opcode]++;
		if (RegB == FREG_31)
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_clear[curr_opcode].counter++;
		    static_float_clear[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Clear",
				curr_opcode, this_PC);
		    break;
		    }
		/* Check_FSQRT will look at zero, one, powers of two */
		AddCallInst(this_inst, InstBefore, "Check_FSQRT", curr_opcode,
				this_PC,
				GetInstRegEnum(this_inst, InstRB));
		break;
		}

	    case MULS_OP:
	    case MULT_OP:
		{
		static_count[curr_opcode]++;
		if ((RegA == FREG_31) || (RegB == FREG_31))
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_clear[curr_opcode].counter++;
		    static_float_clear[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Clear",
				curr_opcode, this_PC);
		    break;
		    }
		AddCallInst(this_inst, InstBefore, "Check_FMUL", curr_opcode,
				this_PC,
				GetInstRegEnum(this_inst, InstRA),
				GetInstRegEnum(this_inst, InstRB));
		break;
		}

	    case CVTLQ_OP:
	    case CVTTQ_OP:
	    case CVTTS_OP:
	    case CVTQL_OP:
	    case CVTST_OP:
		{
		if (RegB == FREG_31)
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_clear[curr_opcode].counter++;
		    static_float_clear[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Clear",
				curr_opcode, this_PC);
		    break;
		    }
		AddCallInst(this_inst, InstBefore, "Check_Float_Zero_Operand",
				curr_opcode,
				this_PC,
				1,	/* Signifying InstRB */
				GetInstRegEnum(this_inst, InstRB));
		break;
		}

	    case DIVS_OP:
	    case DIVT_OP:
		{
		static_count[curr_opcode]++;
		if (RegA == FREG_31)
		    {
		    static_zero_opnd[curr_opcode].counter++;
		    static_zero_opnd[curr_opcode].location = this_PC;

		    static_float_clear[curr_opcode].counter++;
		    static_float_clear[curr_opcode].location = this_PC;
		    AddCallInst(this_inst, InstBefore, "Float_Clear",
				curr_opcode, this_PC);
		    break;
		    }
		AddCallInst(this_inst, InstBefore, "Check_FDIV", curr_opcode,
				this_PC,
				GetInstRegEnum(this_inst, InstRA),
				GetInstRegEnum(this_inst, InstRB));
		}
	    default:
		break;	/* All other opcodes */
	    }	/* switch on relevant opcodes */
	}	/* Each Inst in a block */
      }	/* Each block in a proc */
    }	/* Each proc */
}

void dump_static_array(int this_op)
{
  int   i;

  fprintf(Inst_Logfile, "%s insns:\t%d in total\n",
		Opcode_String[this_op], static_count[this_op]);
  fprintf(Inst_Logfile, "\tShift Distribution\n");
  fprintf(Inst_Logfile, "\tshift\tuse\n");
  fprintf(Inst_Logfile, "\tamount\tcount\n");
}

void InstrumentFini()
{
  int	i;

  fprintf(Inst_Logfile, "*** Static data begin ***\n");
  fprintf(Inst_Logfile, "    Weird FNOPs (using f31 as RegC)\n");
  for (i = 0; i < LAST_OP; i++)
    {
    if (static_weird_FNOP[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%d\t(one such at 0x%lx)\n",
		Opcode_String[i],
		static_weird_FNOP[i].counter,
		static_weird_FNOP[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Zero Operands (uses of f31 as RegA or RegB)\n");
  for (i = 0; i < LAST_OP; i++)
    {
    if (static_zero_opnd[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%d\t(one such at 0x%lx)\n",
		Opcode_String[i],
		static_zero_opnd[i].counter,
		static_zero_opnd[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Float copies (other than cpys) using f31\n");
  for (i = 0; i < LAST_OP; i++)
    {
    if (static_float_copy[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%d\t(one such at 0x%lx)\n",
		Opcode_String[i],
		static_float_copy[i].counter,
		static_float_copy[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Float clears (other than cpys) using f31\n");
  for (i = 0; i < LAST_OP; i++)
    {
    if (static_float_clear[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%d\t(one such at 0x%lx)\n",
		Opcode_String[i],
		static_float_clear[i].counter,
		static_float_clear[i].location);
	}
    }

  fprintf(Inst_Logfile, "    Invariant float branches (f31)\n");
  for (i = 0; i < LAST_OP; i++)
    {
    if (static_invariant_float_branch[i].counter > 0)
	{
	fprintf(Inst_Logfile, "    %s:\t%d\t(one such at 0x%lx)\n",
		Opcode_String[i],
		static_invariant_float_branch[i].counter,
		static_invariant_float_branch[i].location);
	}
    }

	/*
	 * The important data (the rest is basically fluff)
	 */

  fprintf(Inst_Logfile, "    Number MULS uses:\t%d\n", static_count[MULS_OP]);
  fprintf(Inst_Logfile, "    Number MULT uses:\t%d\n", static_count[MULT_OP]);

  fprintf(Inst_Logfile, "    Number DIVS uses:\t%d\n", static_count[DIVS_OP]);
  fprintf(Inst_Logfile, "    Number DIVT uses:\t%d\n", static_count[DIVT_OP]);

  fprintf(Inst_Logfile, "    Number SQRTS uses:\t%d\n", static_count[SQRTS_OP]);
  fprintf(Inst_Logfile, "    Number SQRTT uses:\t%d\n", static_count[SQRTT_OP]);

  fprintf(Inst_Logfile, "*** Static data end ***\n");
  fclose(Inst_Logfile);
  AddCallProto("DynamicFini()");
  AddCallProgram(ProgramAfter, "DynamicFini");
}
