
/*
 *	Helper functions for atom-izing time data collection.
 *	Names/capabilities sort of based upon those in FreePort Express
 */

/*
 * Copyright (C) 2000   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"

/*
 * PrevIntReg - return NULL or pointer to relevant instruction
 */
Inst *PrevIntReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a use of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
/*
 * PrevIntRegUse - return NULL or pointer to relevant instruction
 */
Inst *PrevIntRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * PrevIntRegSet - return NULL or pointer to relevant instruction
 */
Inst *PrevIntRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextIntReg - return NULL or pointer to relevant instruction
 */
Inst *NextIntReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a use of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
/*
 * NextIntRegUse - return NULL or pointer to relevant instruction
 */
Inst *NextIntRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextIntRegUse - return NULL or pointer to relevant instruction
 */

Inst *NextIntRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * PrevFloatReg - return NULL or pointer to relevant instruction
 */
Inst *PrevFloatReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
/*
 * PrevFloatRegUse - return NULL or pointer to relevant instruction
 */
Inst *PrevFloatRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
/*
 * PrevFloatRegSet - return NULL or pointer to relevant instruction
 */
Inst *PrevFloatRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetPrevInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextFloatReg - return NULL or pointer to relevant instruction
 */
Inst *NextFloatReg(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a use of the register */
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;	/* For a set of the register */
    } 
  return NULL; 
}

/*
 * NextFloatRegUse - return NULL or pointer to relevant instruction
 */
Inst *NextFloatRegUse(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.ureg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

/*
 * NextFloatRegSet - return NULL or pointer to relevant instruction
 */
Inst *NextFloatRegSet(Inst *this_inst, RegvType this_reg)
{
  Inst*            temp_inst;
  InstRegUsageVec  this_vec;
  long              mask;

  if (this_inst == NULL)
    return NULL;
  temp_inst = this_inst;
  while ((temp_inst = GetNextInst(temp_inst)) != NULL)
    {
      GetInstRegUsage (temp_inst, &this_vec);   
      mask = 1 << (int) this_reg;
      if ((this_vec.dreg_bitvec[0] & mask) != 0)
        return temp_inst;
    } 
  return NULL; 
}

BOOLEAN SetsIntReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.dreg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

BOOLEAN UsesIntReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.ureg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

BOOLEAN SetsFloatReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.dreg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

BOOLEAN UsesFloatReg(Inst *this_inst, RegvType this_reg)
{
  InstRegUsageVec	this_vec;
  unsigned long		mask;

  GetInstRegUsage (this_inst, &this_vec);
	/* Convert register number into a bit mask */
  mask = 1 << (int) this_reg;

  if ((this_vec.ureg_bitvec[0] & mask) != 0)
    return TRUE;
  return FALSE;
}

enum Opcodes GetOpcode(Inst *this_inst)
{
  int	i;
  int	binhex;

  binhex = GetInstInfo(this_inst, InstBinary);
  for (i=0; i < (int) LAST_OP; i++)
    {
    if ((binhex & OPERATE_MASK) == Opcode_Value[i])
	return (i);
    }
  return (LAST_OP);
}

int GetLitValue(Inst *this_inst)
{
  int	binhex;

  binhex = GetInstInfo(this_inst, InstBinary);
  return (LITERAL_VALUE_(binhex));
}
